# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from trackma import utils


class lib:
    """
    Base interface for creating API implementations for Trackma.

    messenger: Messenger object to send useful messages to
    mediatype: String containing the media type to be used
    """
    name = 'lib'
    version = 'dummy'
    msg = None

    api_info = {'name': 'BaseAPI', 'version': 'undefined', 'merge': False}
    """
    api_info is a dictionary containing useful information about the API itself
    name: API name
    version: API version
    """

    mediatypes = dict()
    """
    mediatypes is a dictionary containing the possible mediatypes for the current API.
    An example mediatype should look like this:
    ::

        mediatypes['anime'] = {
            'has_progress': True,
            'can_add': True,
            'can_delete': True,
            'can_score': True,
            'can_status': True,
            'can_update': True,
            'can_play': True,
            'statuses_start': [1],
            'statuses_finish': [2],
            'statuses':  [1, 2, 3, 4, 6],
            'statuses_dict': { 1: 'Watching', 2: 'Completed', 3: 'On Hold', 4: 'Dropped', 6: 'Plan to Watch' },
        }

    """

    default_mediatype = None

    # Supported signals for the data handler
    signals = {
        'show_info_changed': None,
        'userconfig_changed': None,
    }

    def __init__(self, messenger, account, userconfig):
        """Initializes the API"""
        self.userconfig = userconfig
        self.msg = messenger.with_classname(self.name)
        self.msg.info('Initializing...')

        if not userconfig.get('mediatype'):
            userconfig['mediatype'] = self.default_mediatype

        if userconfig['mediatype'] in self.mediatypes:
            self.mediatype = userconfig['mediatype']
        else:
            raise utils.APIFatal('Unsupported mediatype %s.' %
                                 userconfig['mediatype'])

        self.api_info['mediatype'] = self.mediatype
        self.api_info['supported_mediatypes'] = list(self.mediatypes.keys())

    def _emit_signal(self, signal, *args):
        try:
            if self.signals[signal]:
                self.signals[signal](*args)
        except KeyError:
            raise Exception("Call to undefined signal.")

    def _get_userconfig(self, key):
        return self.userconfig.get(key)

    def _set_userconfig(self, key, value):
        self.userconfig[key] = value

    def connect_signal(self, signal, callback):
        try:
            self.signals[signal] = callback
        except KeyError:
            raise utils.EngineFatal("Invalid signal.")

    def check_credentials(self):
        """Checks if credentials are correct; returns True or False."""
        raise NotImplementedError

    def fetch_list(self):
        """
        Fetches the remote list and returns a dictionary of show dictionaries.

        It should return a dictionary with the show ID as the key and a show dictionary as its value.
        You can create an empty show dictionary with the :func:`utils.show` function.
        """
        raise NotImplementedError

    def add_show(self, item):
        """
        Adds the **item** in the remote server list. The **item** is a show dictionary passed by the Data Handler.
        """
        raise NotImplementedError

    def update_show(self, item):
        """
        Sends the updates of a show to the remote site.

        This function gets called every time a show should be updated remotely,
        and in a queue it may be called many times consecutively, so you should
        use a boolean (or other method) to login only once.

        """
        raise NotImplementedError

    def delete_show(self, item):
        """
        Deletes the **item** in the remote server list. The **item** is a show dictionary passed by the Data Handler.
        """
        raise NotImplementedError

    def search(self, criteria, method):
        """
        Called when the data handler needs a detailed list of shows from the remote server.
        It should return a list of show dictionaries with the additional 'extra' key (which is a list of tuples)
        containing any additional detailed information about the show.
        """
        raise NotImplementedError

    def request_info(self, items):
        # Request detailed information for requested shows
        raise NotImplementedError

    def logout(self):
        # This is called whenever the API won't be required
        # for a good while
        pass

    def media_info(self):
        """Return information about the currently selected mediatype."""
        return self.mediatypes[self.mediatype]

    def set_message_handler(self, message_handler):
        self.msg = message_handler.with_classname(self.name)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import threading

from gi.repository import GLib, GObject, Gtk

from trackma import utils
from trackma.ui.gtk import gtk_dir
from trackma.ui.gtk.showinfobox import ShowInfoBox


class SearchThread(threading.Thread):
    def __init__(self, engine, search_text, callback):
        threading.Thread.__init__(self)
        self._entries = []
        self._error = None
        self._engine = engine
        self._search_text = search_text
        self._callback = callback
        self._stop_request = threading.Event()

    def run(self):
        try:
            self._entries = self._engine.search(self._search_text)
        except utils.TrackmaError as e:
            self._entries = []
            self._error = e

        if not self._stop_request.is_set():
            GLib.idle_add(self._callback, self._entries, self._error)

    def stop(self):
        self._stop_request.set()


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/searchwindow.ui'))
class SearchWindow(Gtk.Window):
    __gtype_name__ = 'SearchWindow'

    __gsignals__ = {
        'search-error': (GObject.SignalFlags.RUN_FIRST, None,
                         (str,))
    }

    btn_add_show = Gtk.Template.Child()
    search_paned = Gtk.Template.Child()
    shows_viewport = Gtk.Template.Child()
    show_info_container = Gtk.Template.Child()
    progress_spinner = Gtk.Template.Child()
    headerbar = Gtk.Template.Child()

    def __init__(self, engine, colors, current_status, transient_for=None):
        Gtk.Window.__init__(self, transient_for=transient_for)
        self.init_template()
        self._entries = []
        self._selected_show = None
        self._showdict = None

        self._engine = engine
        self._current_status = current_status
        self._search_thread = None

        self.showlist = SearchTreeView(colors)
        self.showlist.get_selection().connect("changed", self._on_selection_changed)
        self.showlist.set_size_request(250, 350)
        self.showlist.show()

        self.info = ShowInfoBox(engine, orientation=Gtk.Orientation.VERTICAL)
        self.info.set_size_request(200, 350)
        self.info.show()

        self.shows_viewport.add(self.showlist)
        self.show_info_container.pack_start(self.info, True, True, 0)
        self.search_paned.set_position(400)
        self.set_size_request(450, 350)

    @Gtk.Template.Callback()
    def _on_search_entry_search_changed(self, search_entry):
        search_text = search_entry.get_text().strip()
        self.progress_spinner.start()
        if search_text == "":
            if self._search_thread:
                self._search_thread.stop()
            self._search_finish()
        else:
            self._search(search_text)
            self.progress_spinner.start()

    def _search(self, text):
        if self._search_thread:
            self._search_thread.stop()

        self.headerbar.set_subtitle("Searching: \"%s\"" % text)
        self._search_thread = SearchThread(self._engine,
                                           text,
                                           self._search_finish_idle)
        self._search_thread.start()

    def _search_finish(self):
        self.headerbar.set_subtitle(
            "%s result%s." % ((len(self._entries), 's')
                              if len(self._entries) > 0
                              else ('No', '')
                              )
        )
        self.progress_spinner.stop()

    def _search_finish_idle(self, entries, error):
        self._entries = entries
        self._showdict = dict()
        self._search_finish()
        self.showlist.append_start()
        for show in entries:
            self._showdict[show['id']] = show
            self.showlist.append(show)
        self.showlist.append_finish()

        self.btn_add_show.set_sensitive(False)

        if error:
            self.emit('search-error', error)

    @Gtk.Template.Callback()
    def _on_btn_add_show_clicked(self, btn):
        show = self._get_full_selected_show()

        if show is not None:
            self._add_show(show)

    def _get_full_selected_show(self):
        for item in self._entries:
            if item['id'] == self._selected_show:
                return item

        return None

    def _add_show(self, show):
        try:
            self._engine.add_show(show, self._current_status)
        except utils.TrackmaError as e:
            self.emit('search-error', e)

    def _on_selection_changed(self, selection):
        # Get selected show ID
        (tree_model, tree_iter) = selection.get_selected()
        if not tree_iter:
            return

        self._selected_show = int(tree_model.get(tree_iter, 0)[0])
        if self._selected_show in self._showdict:
            self.info.load(self._showdict[self._selected_show])
            self.btn_add_show.set_sensitive(True)


class SearchTreeView(Gtk.TreeView):
    def __init__(self, colors):
        Gtk.TreeView.__init__(self)

        self.cols = dict()
        i = 1
        for name in ('Title', 'Type', 'Total'):
            self.cols[name] = Gtk.TreeViewColumn(name)
            self.cols[name].set_sort_column_id(i)
            self.append_column(self.cols[name])
            i += 1

        # renderer_id = Gtk.CellRendererText()
        # self.cols['ID'].pack_start(renderer_id, False)
        # self.cols['ID'].add_attribute(renderer_id, 'text', 0)

        renderer_title = Gtk.CellRendererText()
        self.cols['Title'].pack_start(renderer_title, False)
        self.cols['Title'].set_resizable(True)
        self.cols['Title'].set_expand(False)
        self.cols['Title'].add_attribute(renderer_title, 'text', 1)
        self.cols['Title'].add_attribute(renderer_title, 'foreground', 4)

        renderer_type = Gtk.CellRendererText()
        self.cols['Type'].pack_start(renderer_type, False)
        self.cols['Type'].add_attribute(renderer_type, 'text', 2)

        renderer_total = Gtk.CellRendererText()
        self.cols['Total'].pack_start(renderer_total, False)
        self.cols['Total'].add_attribute(renderer_total, 'text', 3)

        self.store = Gtk.ListStore(str, str, str, str, str)
        self.set_model(self.store)

        self.colors = colors

    def append_start(self):
        self.freeze_child_notify()
        self.store.clear()

    def append(self, show):
        if show['status'] == 1:
            color = self.colors['is_airing']
        elif show['status'] == 3:
            color = self.colors['not_aired']
        else:
            color = None

        row = [
            str(show['id']),
            str(show['title']),
            str(show['type']),
            str(show['total']),
            color]
        self.store.append(row)

    def append_finish(self):
        self.thaw_child_notify()
        self.store.set_sort_column_id(1, Gtk.SortType.ASCENDING)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess
import time

from trackma import utils
from trackma.tracker import tracker


class PollingTracker(tracker.TrackerBase):
    name = 'Tracker (polling)'

    def get_playing_file(self, watch_dirs, players):
        for path in watch_dirs:
            # TODO: We'll run lsof once for each directory for now.
            try:
                lsof = subprocess.Popen(
                    ['lsof', '-w', '-n', '-c', ''.join(['/', players, '/']), '-Fn', path], stdout=subprocess.PIPE)
            except OSError:
                self.msg.warn("Couldn't execute lsof. Disabling tracker.")
                self.disable()
                return None

            output = lsof.communicate()[0].decode('utf-8')

            for line in output.splitlines():
                if line[0] == 'n' and utils.is_media(line):
                    return os.path.basename(line[1:])

        return None

    def observe(self, config, watch_dirs):
        self.msg.info("pyinotify not available; using polling (slow).")
        while self.active:
            # This runs the tracker and update the playing show if necessary
            filename = self.get_playing_file(
                watch_dirs, config['tracker_process'])
            (state, show_tuple) = self._get_playing_show(filename)
            self.update_show_if_needed(state, show_tuple)

            # Wait for the interval before running check again
            time.sleep(config['tracker_interval'])



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import json
import socket
import ssl

from trackma import utils
from trackma.lib.lib import lib


class libvndb(lib):
    """
    API class to communicate with VNDB
    Implements protocol version 1.

    Website: https://www.vndb.org
    API documentation: https://vndb.org/d11
    Designed by: Yorhel <contact@vndb.org> (https://vndb.org/u2)

    """
    name = 'libvndb'

    api_info = {
        'name': 'VNDB',
        'shortname': 'vndb',
        'version': 2,
        'merge': True,
    }

    default_mediatype = 'vnlist'
    pagesize_list = 100
    pagesize_details = 25

    mediatypes = dict()
    mediatypes['vnlist'] = {
        'has_progress': False,
        'can_score': True,
        'can_status': True,
        'can_add': True,
        'can_delete': True,
        'can_update': False,
        'can_play': False,
        'statuses':  [1, 2, 3, 4, 0],
        'statuses_dict': {1: 'Playing', 2: 'Finished', 3: 'Stalled', 4: 'Dropped', 0: 'Unknown'},
        'score_max': 10,
        'score_step': 0.1,
        'statuses_start': [1],
        'statuses_finish': [2],
    }
    mediatypes['wishlist'] = {
        'has_progress': False,
        'can_score': True,
        'can_status': True,
        'can_add': True,
        'can_delete': True,
        'can_update': False,
        'can_play': False,
        'statuses':  [0, 1, 2, 3],
        'statuses_dict': {0: 'High', 1: 'Medium', 2: 'Low', 3: 'Blacklist'},
        'score_max': 10,
        'score_step': 0.1,
        'statuses_start': [],
        'statuses_finish': [],
    }

    def __init__(self, messenger, account, userconfig):
        """Initializes the useragent through credentials."""
        super(libvndb, self).__init__(messenger, account, userconfig)

        self.hostname = "api.vndb.org"
        self.tls = 19535
        self.tcp = 19534
        self.username = account['username']
        self.password = account['password']
        self.logged_in = False

    def _connect(self):
        """Create TCP socket and connect"""

        self.context = ssl.create_default_context()
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.s = self.context.wrap_socket(s, server_hostname=self.hostname)
            self.s.connect((self.hostname, self.tls))
        except socket.error:
            raise utils.APIError("Connection error.")
        except ssl.CertificateError:
            raise utils.APIError("Insecure connection.")

    def _disconnect(self):
        """Shutdown and close the socket"""
        self.s.shutdown(socket.SHUT_RDWR)
        self.s.close()

    def _sendcmd(self, cmd, options=None):
        """Send a VNDB compatible command and return the response data"""
        msg = cmd
        if options:
            msg += " " + json.dumps(options, separators=(',', ':'))
        msg = msg.encode('utf-8')
        msg += b"\x04"  # EOT

        # Send message
        self.s.sendall(msg)

        # Construct response
        lines = []
        while True:
            line = self.s.recv(65536)
            if line.endswith(b"\x04"):
                line = line.strip(b"\x04")
                lines.append(line)
                response = b"".join(lines).decode('utf-8')
                break
            else:
                lines.append(line)

        # Separate into response name and JSON data
        _resp = response.split(' ', 1)
        name = _resp[0]
        try:
            data = json.loads(_resp[1])
        except IndexError:
            data = None

        # Treat error as an exception
        if name == 'error':
            raise utils.APIError(data['msg'])

        return (name, data)

    def check_credentials(self):
        """Checks if credentials are correct; returns True or False."""
        if self.logged_in:
            return True

        self.msg.info('Connecting...')
        self._connect()

        self.msg.info('Logging in...')
        (name, data) = self._sendcmd('login',
                                     {'protocol': 1,
                                      'client': 'Trackma',
                                      'clientver': self.api_info['version'],
                                      'username': self.username,
                                      'password': self.password,
                                      })

        if name == 'ok':
            self.logged_in = True
            self._set_userconfig('username', self.username)
            return True
        else:
            return False

    def fetch_list(self):
        """Queries the full list from the remote server.
        Returns the list if successful, False otherwise."""
        self.check_credentials()

        # Retrieve VNs per pages
        page = 1
        vns = dict()
        while True:
            self.msg.info('Downloading list... (%d)' % page)

            (name, data) = self._sendcmd('get %s basic (uid = 0)' % self.mediatype,
                                         {'page': page,
                                          'results': self.pagesize_list
                                          })

            # Something is wrong if we don't get a results response.
            if name != 'results':
                raise utils.APIFatal("Invalid response (%s)" % name)

            # Process list
            for item in data['items']:
                vnid = item['vn']
                vns[vnid] = utils.show()
                vns[vnid]['id'] = vnid
                vns[vnid]['url'] = self._get_url(vnid)
                vns[vnid]['my_status'] = item.get(
                    'status', item.get('priority'))

            if not data['more']:
                # No more VNs, finish
                break
            page += 1

        # Retrieve scores per pages
        page = 1
        while True:
            self.msg.info('Downloading votes... (%d)' % page)

            (name, data) = self._sendcmd('get votelist basic (uid = 0)',
                                         {'page': page,
                                          'results': self.pagesize_list
                                          })

            # Something is wrong if we don't get a results response.
            if name != 'results':
                raise utils.APIFatal("Invalid response (%s)" % name)

            for item in data['items']:
                vnid = item['vn']
                if vnid not in vns:
                    # Ghost vote; create entry for it.
                    vns[vnid] = utils.show()
                    vns[vnid]['id'] = vnid
                    vns[vnid]['url'] = self._get_url(vnid)
                    vns[vnid]['my_status'] = 0

                vns[vnid]['my_score'] = (item['vote'] / 10.0)
                vns[vnid]['my_finish_date'] = datetime.datetime.fromtimestamp(
                    item['added'])

            if not data['more']:
                # No more VNs, finish
                break
            page += 1

        return vns

    def request_info(self, itemlist):
        self.check_credentials()

        start = 0
        infos = list()
        remaining = [show['id'] for show in itemlist]
        while True:
            self.msg.info('Requesting details...(%d)' % start)
            end = start + self.pagesize_details

            (name, data) = self._sendcmd('get vn basic,details (id = %s)' % repr(remaining[start:end]),
                                         {'page': 1,
                                          'results': self.pagesize_details,
                                          })

            # Something is wrong if we don't get a results response.
            if name != 'results':
                raise utils.APIFatal("Invalid response (%s)" % name)

            # Process list
            for item in data['items']:
                infos.append(self._parse_info(item))

            start += self.pagesize_details
            if start >= len(itemlist):
                # We're going beyond the list, finish
                break

        self._emit_signal('show_info_changed', infos)
        return infos

    def add_show(self, item):
        # When we try to "update" a VN that isn't in the list, the
        # VNDB API automatically inserts it, so we can use the same
        # command.
        self.update_show(item)

    def update_show(self, item):
        self.check_credentials()

        # Update status with set vnlist
        if 'my_status' in item:
            self.msg.info('Updating VN %s (status)...' %
                          item['title'])

            if self.mediatype == 'wishlist':
                values = {'priority': item['my_status']}
            else:
                values = {'status': item['my_status']}

            (name, data) = self._sendcmd('set %s %d' %
                                         (self.mediatype, item['id']), values)

            if name != 'ok':
                raise utils.APIError("Invalid response (%s)" % name)

        # Update vote with set votelist
        if 'my_score' in item:
            self.msg.info('Updating VN %s (vote)...' %
                          item['title'])

            if item['my_score'] > 0:
                # Add or update vote
                values = {'vote': item['my_score'] * 10}
            else:
                # Delete vote if it's 0
                values = None

            (name, data) = self._sendcmd(
                'set votelist %d' % item['id'], values)

            if name != 'ok':
                raise utils.APIError("Invalid response (%s)" % name)

    def delete_show(self, item):
        self.check_credentials()

        self.msg.info('Deleting VN %s...' % item['title'])

        (name, data) = self._sendcmd('set %s %d' %
                                     (self.mediatype, item['id']))

        if name != 'ok':
            raise utils.APIError("Invalid response (%s)" % name)

    def search(self, criteria, method):
        self.check_credentials()

        results = list()
        self.msg.info('Searching for %s...' % criteria)

        (name, data) = self._sendcmd('get vn basic,details (search ~ "%s")' % criteria,
                                     {'page': 1,
                                      'results': self.pagesize_details,
                                      })

        # Something is wrong if we don't get a results response.
        if name != 'results':
            raise utils.APIFatal("Invalid response (%s)" % name)

        # Process list
        for item in data['items']:
            results.append(self._parse_info(item))

        self._emit_signal('show_info_changed', results)

        if not results:
            raise utils.APIError('No results.')

        return results

    def logout(self):
        self.msg.info('Disconnecting...')
        self._disconnect()
        self.logged_in = False

    def merge(self, show, info):
        show['title'] = info['title']
        show['image'] = info.get('image')
        show['start_date'] = info.get('start_date')
        if show['start_date'] and show['start_date'] > datetime.datetime.now():
            show['status'] = utils.Status.NOTYET
        else:
            show['status'] = utils.Status.FINISHED

    def _parse_info(self, item):
        start_date = self._str2date(item['released'])

        info = utils.show()
        info.update({'id': item['id'],
                     'title': item['title'],
                     'image': item['image'],
                     'url': self._get_url(item['id']),
                     'start_date': self._str2date(item['released']),
                     'extra': [
                    ('Original Name', item['original']),
                    ('Released',      item['released']),
                    ('Languages',     ','.join(item['languages'])),
                    ('Original Language', ','.join(item['orig_lang'])),
                    ('Platforms',     ','.join(item['platforms'])),
                    ('Aliases',       item['aliases']),
                    ('Length',        item['length']),
                    ('Description',   item['description']),
                    ('Links',         item['links']),
                    ]
        })
        return info

    def _get_url(self, vnid):
        return "https://vndb.org/v%d" % vnid

    def _str2date(self, string):
        if string != '0000-00-00':
            try:
                return datetime.datetime.strptime(string, "%Y-%m-%d")
            except ValueError:
                return None  # Ignore date if it's invalid
        else:
            return None



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from PyQt6 import QtCore, QtGui
from PyQt6.QtWidgets import (QAbstractItemView, QGridLayout, QHeaderView, QLabel, QListView, QScrollArea, QSplitter,
                             QTableView, QVBoxLayout, QWidget)

from trackma import utils
from trackma.ui.qt.delegates import AddListDelegate, ShowsTableDelegate
from trackma.ui.qt.models import AddListModel, AddListProxy, AddTableModel, ShowListModel, ShowListProxy
from trackma.ui.qt.workers import ImageWorker


class DetailsWidget(QWidget):
    def __init__(self, parent, worker):
        self.worker = worker

        QWidget.__init__(self, parent)

        # Build layout
        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        self.show_title = QLabel()
        show_title_font = QtGui.QFont()
        show_title_font.setBold(True)
        show_title_font.setPointSize(12)
        self.show_title.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.show_title.setFont(show_title_font)

        info_area = QWidget()
        info_layout = QGridLayout()

        self.show_image = QLabel()
        self.show_image.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.show_info = QLabel()
        self.show_info.setWordWrap(True)
        self.show_info.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.show_description = QLabel()
        self.show_description.setWordWrap(True)
        self.show_description.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        info_layout.addWidget(self.show_image,        0, 0, 1, 1)
        info_layout.addWidget(self.show_info,         1, 0, 1, 1)
        info_layout.addWidget(self.show_description,  0, 1, 2, 1)

        info_area.setLayout(info_layout)

        scroll_area = QScrollArea()
        scroll_area.setBackgroundRole(QtGui.QPalette.ColorRole.Light)
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(info_area)

        main_layout.addWidget(self.show_title)
        main_layout.addWidget(scroll_area)

        self.setLayout(main_layout)

    def worker_call(self, function, ret_function, *args, **kwargs):
        # Run worker in a thread
        self.worker.set_function(function, ret_function, *args, **kwargs)
        self.worker.start()

    def load(self, show):
        metrics = QtGui.QFontMetrics(self.show_title.font())
        title = metrics.elidedText(
            show['title'], QtCore.Qt.TextElideMode.ElideRight, self.show_title.width())

        self.show_title.setText("<a href=\"%s\">%s</a>" % (show['url'], title))
        self.show_title.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self.show_title.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextBrowserInteraction)
        self.show_title.setOpenExternalLinks(True)

        # Load show info
        self.show_info.setText('Wait...')
        self.worker_call('get_show_details', self.r_details_loaded, show)
        api_info = self.worker.engine.api_info

        # Load show image
        if show.get('image'):
            utils.make_dir(utils.to_cache_path())
            filename = utils.to_cache_path("%s_%s_f_%s.jpg" % (
                api_info['shortname'], api_info['mediatype'], show['id']))

            if os.path.isfile(filename):
                self.s_show_image(filename)
            else:
                self.show_image.setText('Downloading...')
                self.image_worker = ImageWorker(
                    show['image'], filename, (200, 280))
                self.image_worker.finished.connect(self.s_show_image)
                self.image_worker.start()
        else:
            self.show_image.setText('No image')

    def s_show_image(self, filename):
        self.show_image.setPixmap(QtGui.QPixmap(filename))

    def r_details_loaded(self, result):
        if result['success']:
            details = result['result']

            info_strings = []
            description_strings = []
            # This might come down to personal preference
            description_keys = {'Synopsis', 'English', 'Japanese',  'Romaji', 'Synonyms'}

            for key, value in details['extra']:
                if not key or not value:
                    continue
                str_value = ', '.join(value) if isinstance(value, list) else str(value)
                if key in description_keys or len(str_value) >= 17:
                    # Avoid short tidbits taking up too much vertical space
                    html = f"<h3>{key}</h3><p>{str_value}</p>"
                else:
                    html = f"<p><b>{key}:</b> {str_value}</p>"

                if key in description_keys:
                    description_strings.append(html)
                else:
                    info_strings.append(html)

            info_string = ''.join(info_strings)
            self.show_info.setText(info_string)
            description_string = ''.join(description_strings)
            self.show_description.setText(description_string)
        else:
            self.show_info.setText('There was an error while getting details.')


class ShowsTableView(QTableView):
    """
    Regular table widget with context menu for show actions.

    """
    middleClicked = QtCore.pyqtSignal()

    def __init__(self, parent=None, palette=None):
        QTableView.__init__(self, parent)

        model = ShowListModel(palette=palette)
        proxymodel = ShowListProxy()
        proxymodel.setSourceModel(model)
        proxymodel.setFilterKeyColumn(-1)
        self.setModel(proxymodel)

        self.setItemDelegate(ShowsTableDelegate(self, palette=palette))
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.horizontalHeader().setHighlightSections(False)
        self.horizontalHeader().setSectionsMovable(True)
        self.horizontalHeader().setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.verticalHeader().hide()
        self.setGridStyle(QtCore.Qt.PenStyle.NoPen)

    def contextMenuEvent(self, event):
        action = self.context_menu.exec(event.globalPos())

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

        if event.button() == QtCore.Qt.MouseButton.MiddleButton:
            self.middleClicked.emit()


class AddCardView(QListView):
    changed = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None, api_info=None):
        super().__init__(parent)

        m = AddListModel(api_info=api_info)
        proxy = AddListProxy()
        proxy.setSourceModel(m)
        proxy.sort(0, QtCore.Qt.SortOrder.AscendingOrder)

        self.setItemDelegate(AddListDelegate())
        self.setFlow(QListView.Flow.LeftToRight)
        self.setWrapping(True)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setModel(proxy)

        self.selectionModel().currentRowChanged.connect(self.s_show_selected)

    def s_show_selected(self, new, old=None):
        if not new:
            return

        index = self.model().mapToSource(new).row()
        selected_show = self.getModel().results[index]

        self.changed.emit(selected_show)

    def setResults(self, results):
        self.getModel().setResults(results)

    def getModel(self):
        return self.model().sourceModel()


class AddTableDetailsView(QSplitter):
    """ This is a splitter widget that contains a table and
    a details widget. Used in the Add Show dialog. """

    changed = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None, worker=None):
        super().__init__(parent)

        self.table = QTableView()
        m = AddTableModel()
        proxy = QtCore.QSortFilterProxyModel()
        proxy.setSourceModel(m)

        self.table.setGridStyle(QtCore.Qt.PenStyle.NoPen)
        self.table.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.table.setModel(proxy)

        # Allow sorting but don't sort by default
        self.table.horizontalHeader().setSortIndicator(-1, QtCore.Qt.SortOrder.AscendingOrder)
        self.table.setSortingEnabled(True)

        self.table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)

        self.table.selectionModel().currentRowChanged.connect(self.s_show_selected)
        self.addWidget(self.table)

        self.details = DetailsWidget(parent, worker)
        self.addWidget(self.details)

        self.setSizes([1, 1])

    def s_show_selected(self, new, old=None):
        if not new:
            return

        index = self.table.model().mapToSource(new).row()
        selected_show = self.getModel().results[index]
        self.details.load(selected_show)

        self.changed.emit(selected_show)

    def setResults(self, results):
        self.getModel().setResults(results)

    def getModel(self):
        return self.table.model().sourceModel()

    def clearSelection(self):
        return self.table.clearSelection()
